/***************************************************************************
                  hwsensors.c - eboxy lm_sensors plugin
                             -------------------
    begin                : Mon Jan 6 2003
    copyright            : (C) 2003 by Paul Eggleton
    email                : bluelightning@bluelightning.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   Code in this module is taken from the prog/sensors/main.c & chips.c   *
 *   from lm_sensors 2.70.                                                 *
 *                                                                         *
 ***************************************************************************/

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>

#include "eboxyplugin.h"
#include "pluginconstants.h"
#include "linux/sensors.h"
#include "sensors/sensors.h"
#include "sensors/error.h"

#define DEFAULT_CONFIG_FILE_NAME "sensors.conf"

/* Globals */
static char *config_file_name;
FILE *config_file;
static const char *config_file_path[] =
  { "/etc", "/usr/local/etc", "/usr/lib/sensors", "/usr/local/lib/sensors",
    "/usr/lib", "/usr/local/lib", ".", 0 };

/* Function prototypes */
void print_unknown_chip(const sensors_chip_name *name);
static void open_config_file(void);
static int open_this_config_file(char *filename);
int sensors_get_label_and_valid(sensors_chip_name name, int feature, char **label, int *valid);
char *hws_readsensor(const char *sender, int numargs, const char *args[]);

/* Implementation */

int ebplugin_init(void) {
  int rc = 0;

  /* Set version info for the plugin */
  setPluginInfo("Sensors", "0.1");

  /* Register an object to be used in scripts */
  rc = registerObject("hwsensors");
  if(rc)
    return rc;
  rc = registerMethodDL("hwsensors", "readsensor", 2, "hws_readsensor");
  if(rc)
    return rc;

  open_config_file();

  if((rc = sensors_init(config_file))) {
    fprintf(stderr,"%s\n",sensors_strerror(rc));
    if(rc == -SENSORS_ERR_PROC)
      fprintf(stderr,
              "hwsensors: /proc/sys/dev/sensors/chips or /proc/bus/i2c unreadable;\n"
              "hwsensors: Make sure you have done 'modprobe i2c-proc'!\n");
    return 1;
  }

  /* Success */
  return 0;
}

int ebplugin_message(int msgcode, void *msgdata) {
  return 0;     /* return 0 here for future compatibility */
}

void ebplugin_deinit(void) {
  /* No deinitialisation to be done */
}

char *hws_readsensor(const char *sender, int numargs, const char *args[]) {
  sensors_chip_name chip;
  int a,b,valid,rc;
  const sensors_feature_data *data;
  char *label;
  double val;
  char *buf;

  rc = sensors_parse_chip_name(args[0], &chip);
  if(rc==0) {

    a=b=0;
    while((data=sensors_get_all_features(chip, &a, &b))) {
      if (sensors_get_label_and_valid(chip, data->number, &label, &valid)) {
        printf("hwsensors: ERROR: Can't get feature `%s' data!\n", data->name);
        continue;
      }
      if(!valid)
        continue;
      if (data->mode & SENSORS_MODE_R) {
        if(sensors_get_feature(chip, data->number, &val)) {
          printf("hwsensors: ERROR: Can't get feature `%s' data!\n", data->name);
          continue;
        }
/*        if (data->mapping != SENSORS_NO_MAPPING)
          printf("  %s: %.2f (%s)\n", label, val, data->name);
        else
          printf("%s: %.2f (%s)\n", label, val, data->name);*/
        if(!strcmp(data->name, args[1])) {
          buf = malloc(16);
          snprintf(buf, 16, "%.9g", val);
          return buf;
        }

      }
/*      else
        printf("(%s)",label);*/
    }


  }
  else
    fprintf(stderr, "hwsensors: couldn't get chip %s (code %d)\n", args[0], rc);
  return NULL;
}

int sensors_get_label_and_valid(sensors_chip_name name, int feature, char **label, int *valid) {
  int err;
  err = sensors_get_label(name,feature,label);
  if (!err)
    err = sensors_get_ignored(name,feature);
  if (err >= 0) {
    *valid = err;
    err = 0;
  }
  return err;
}

/* This examines global var config_file, and leaves the name there too.
   It also opens config_file. */
void open_config_file(void) {
#define MAX_FILENAME_LEN 1024
  char *filename;
  char buffer[MAX_FILENAME_LEN];
  int res,i;

  if(config_file_name && !strcmp(config_file_name,"-")) {
    config_file = stdin;
    return;
  } else if (config_file_name && index(config_file_name,'/')) {
    if ((res = open_this_config_file(config_file_name))) {
      fprintf(stderr,"hwsensors: Could not locate or open config file\n");
      fprintf(stderr,"hwsensors: %s: %s\n",config_file_name,strerror(res));
      return;
    }
  }
  else {
    if (config_file_name)
      filename = config_file_name;
    else
      filename = strdup(DEFAULT_CONFIG_FILE_NAME);
    for (i = 0; config_file_path[i]; i++) {
      if ((snprintf(buffer,MAX_FILENAME_LEN,
                   "%s/%s",config_file_path[i],filename)) < 1) {
        fprintf(stderr,
                "hwsensors: open_config_file: ridiculous long config file name!\n");
        return;
      }
      if (!open_this_config_file(buffer)) {
        free(config_file_name);
        config_file_name = strdup(buffer);
        return;
      }
    }
    fprintf(stderr,"hwsensors: Could not locate or open config file!\n");
    return;
  }
}

int open_this_config_file(char *filename) {
  config_file = fopen(filename,"r");
  if(!config_file)
    return -errno;
  return 0;
}
