/***************************************************************************
                      mysql.c - eboxy MySQL plugin
                             -------------------
    begin                : Mon Jan 13 2003
    copyright            : (C) 2003 by Paul Eggleton <bluelightning@bluelightning.org>
                                   and William Stotts <wstotts@snet.net>
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>

#include "eboxyplugin.h"
#include "pluginconstants.h"
#include "mysql.h"
#include "mysql_version.h"

/* Types */
struct llist {
  struct llistitem *first;
  struct llistitem *last;
  unsigned int count;
};
struct llistitem {
  char *name;
  char *value;
  struct llistitem *next;
};

/* Globals */
MYSQL *mysqlobj;
MYSQL_RES *res_set;
MYSQL_FIELD *fields;
MYSQL_ROW currentrow;
char *hostname;
int port;
int bindingenabled;
int rowindex;
struct llist boundlist;

/* Function prototypes */
char *msp_connect(const char *sender, int numargs, const char *args[]);
char *msp_gethostname(const char *sender);
int msp_sethostname(const char *sender, const char *value);
char *msp_query(const char *sender, int numargs, const char *args[]);
char *msp_getrows(const char *sender);
char *msp_next(const char *sender, int numargs, const char *args[]);
char *msp_previous(const char *sender, int numargs, const char *args[]);
char *msp_first(const char *sender, int numargs, const char *args[]);
char *msp_fieldbyname(const char *sender, int numargs, const char *args[]);
char *msp_bindwidget(const char *sender, int numargs, const char *args[]);
char *msp_unbindwidget(const char *sender, int numargs, const char *args[]);
char *msp_getbindingenabled(const char *sender);
int msp_setbindingenabled(const char *sender, const char *value);

void update_bound_widgets(void);
void on_scroll(void);
void set_widget_value(const char *name, const char *value);

char *inttostr(int n);
int strtobool(const char *str);
void llist_add(struct llist *list, const char *name, const char *value);
const char *llist_get(struct llist *list, int index);
void llist_delete(struct llist *list, int index);
void llist_clear(struct llist *list);
int llist_find(struct llist *list, const char *name);

/* Implementation */

int ebplugin_init(void) {
  int rc = 0;

  /* Set version info for the plugin */
  setPluginInfo("MySQLQuery", "0.2");

  /* Register an object to be used in scripts */
  rc = registerObject("mysql");
  if(rc)
    return rc;
  rc = registerMethodDL("mysql", "connect", 3, "msp_connect");
  if(rc)
    return rc;
  rc = registerMethodDL("mysql", "query", 1, "msp_query");
  if(rc)
    return rc;
  rc = registerPropertyDL("mysql", "hostname", "msp_gethostname", "msp_sethostname");
  if(rc)
    return rc;
  rc = registerPropertyDL("mysql", "port", "msp_getport", "msp_setport");
  if(rc)
    return rc;
  rc = registerPropertyDL("mysql", "rows", "msp_getrows", NULL);
  if(rc)
    return rc;
  rc = registerMethodDL("mysql", "next", 0, "msp_next");
  if(rc)
    return rc;
  rc = registerMethodDL("mysql", "previous", 0, "msp_previous");
  if(rc)
    return rc;
  rc = registerMethodDL("mysql", "first", 0, "msp_first");
  if(rc)
    return rc;
  rc = registerMethodDL("mysql", "fieldbyname", 1, "msp_fieldbyname");
  if(rc)
    return rc;
  rc = registerMethodDL("mysql", "bindwidget", 2, "msp_bindwidget");
  if(rc)
    return rc;
  rc = registerMethodDL("mysql", "unbindwidget", 1, "msp_unbindwidget");
  if(rc)
    return rc;
  rc = registerPropertyDL("mysql", "bindingenabled", "msp_getbindingenabled", "msp_setbindingenabled");
  if(rc)
    return rc;

  mysqlobj = mysql_init(NULL);
  if(mysqlobj == NULL) {
    fprintf(stderr, "mysql: failed to initialise MySQL client library (out of memory?)\n");
    return 1;
  }

  /* Default values */
  hostname = strdup("localhost");
  port = MYSQL_PORT;
  bindingenabled = 1;
  boundlist.first = NULL;
  boundlist.last = NULL;
  boundlist.count = 0;

  res_set = NULL;
  currentrow = NULL;
  rowindex = 0;

  /* Success */
  return 0;
}

int ebplugin_message(int msgcode, void *msgdata) {
  return 0;     /* return 0 here for future compatibility */
}

void ebplugin_deinit(void) {
  if(res_set)
    mysql_free_result(res_set);
  mysql_close(mysqlobj);
  free(hostname);
  llist_clear(&boundlist);
}

char *msp_gethostname(const char *sender) {
  return strdup(hostname);
}

int msp_sethostname(const char *sender, const char *value) {
  free(hostname);
  hostname = strdup(value);
  return 0;
}

char *msp_getport(const char *sender) {
  return inttostr(port);
}

int msp_setport(const char *sender, const char *value) {
  int portnum;
  portnum = atoi(value);
  if(portnum > 0 && portnum < 65536) {
    port = portnum;
    return 0;
  }
  else
    return -1;
}

char *msp_connect(const char *sender, int numargs, const char *args[]) {
  mysql_options(mysqlobj, MYSQL_READ_DEFAULT_GROUP, "eboxy");
  if(!mysql_real_connect(mysqlobj, hostname, args[0], args[1], args[2], port, NULL, CLIENT_FOUND_ROWS)) {
    const char *errstring = mysql_error(mysqlobj);
    fprintf(stderr, "mysql: failed to connect to database: Error: %s\n", errstring);
    return strdup(errstring);
  }

  return NULL;
}

char *msp_query(const char *sender, int numargs, const char *args[]) {
  int rc;
  
  if(mysql_query(mysqlobj, args[0]) != 0) {
    const char *errstring = mysql_error(mysqlobj);
    fprintf(stderr, "mysql: query statement failed: Error: %s\n", errstring);
    return strdup(errstring);
  }
  else {
    rowindex = 0;
    currentrow = NULL;
    
    if(res_set) {
      /* A query has been run already, free the result set */
      mysql_free_result(res_set);
      res_set = NULL;
    }

    /* Get result set */
    res_set = mysql_store_result(mysqlobj);
    if(res_set == NULL) {
      if (mysql_field_count(mysqlobj) > 0)
        fprintf(stderr, "mysql: Error processing result set\n");
      else {
        /* No results (an INSERT query, etc.)*/
      }
    }
    else {
      printf("mysql: query succeeded: %lu rows affected\n",
              (unsigned long) mysql_affected_rows (mysqlobj));
      /* Get field data */
      fields = mysql_fetch_fields(res_set);
      /* Fetch first row */
      currentrow = mysql_fetch_row(res_set);
      rc = mysql_errno(mysqlobj);
      if(rc!=0)
        fprintf(stderr, "mysql: mysql_fetch_row() failed (error code %d)\n", rc);
      else {
        on_scroll();
      }
    }
  }

  return NULL;
}

char *msp_getrows(const char *sender) {
  int rowcount;

  if(res_set)
    rowcount = mysql_num_rows(res_set);
  else {
    printf("mysql: fieldbyname: no result set\n");
    rowcount = 0;
  }

  return inttostr(rowcount);
}

char *msp_next(const char *sender, int numargs, const char *args[]) {
  int rc;

  if(res_set) {
    if(currentrow != NULL) {
      /* Get next row */
      currentrow = mysql_fetch_row(res_set);

      if(currentrow == NULL) {
        rc = mysql_errno(mysqlobj);
        if(rc!=0) {
          fprintf(stderr, "mysql: mysql_fetch_row() failed (error code %d)\n", rc);
          rowindex = 0;
        }
        else {
          return strdup("EOF"); /* No more records */
        }
      }
      else {
        rowindex++;
        on_scroll();
      }
    }
    else
      printf("mysql: already at EOF\n");
  }
  else
    fprintf(stderr, "mysql: next: no result set\n");

  return NULL;
}

char *msp_previous(const char *sender, int numargs, const char *args[]) {
  int rc;

  if(res_set) {
    
    if(rowindex > 0) {
      /* Get previous row */
      mysql_data_seek(res_set, rowindex-1);
      currentrow = mysql_fetch_row(res_set);

      if(currentrow == NULL) {
        rc = mysql_errno(mysqlobj);
        if(rc!=0)
          fprintf(stderr, "mysql: mysql_fetch_row() failed (error code %d)\n", rc);
        rowindex = 0;
      }
      else {
        rowindex--;
        on_scroll();
      }
    }
    else
      printf("mysql: already at first record\n");
  }
  else
    fprintf(stderr, "mysql: next: no result set\n");

  return NULL;
}

char *msp_first(const char *sender, int numargs, const char *args[]) {
  int rc;

  if(res_set) {
    rowindex = 0;
    mysql_data_seek(res_set, 0);
    currentrow = mysql_fetch_row(res_set);

    if(currentrow == NULL) {
      rc = mysql_errno(mysqlobj);
      if(rc!=0)
        fprintf(stderr, "mysql: mysql_fetch_row() failed (error code %d)\n", rc);
    }
    else {
      on_scroll();
    }
  }
  else
    fprintf(stderr, "mysql: first: no result set\n");

  return NULL;
}

char *msp_fieldbyname(const char *sender, int numargs, const char *args[]) {
  int i, numfields;

  if(res_set) {
    numfields = mysql_num_fields(res_set);

    for(i=0; i < numfields; i++) {
      if(strcasecmp(args[0], fields[i].name) == 0)
        return strdup(currentrow[i]);
    }

    fprintf(stderr, "mysql: no field named %s in result set\n", args[0]);
  }
  else
    fprintf(stderr, "mysql: fieldbyname: no result set\n");

  return NULL;
}

char *msp_bindwidget(const char *sender, int numargs, const char *args[]) {
  int index;

  index = llist_find(&boundlist, args[0]);
  if(index>-1) {
    fprintf(stderr, "mysql: widget %s is already bound\n", args[0]);
  }
  else {
    llist_add(&boundlist, args[0], args[1]);
    if(bindingenabled == 1 && res_set != NULL && boundlist.count > 0)
      update_bound_widgets();
  }

  return NULL;
}

char *msp_unbindwidget(const char *sender, int numargs, const char *args[]) {
  int index;

  index = llist_find(&boundlist, args[0]);
  if(index>-1) {
    llist_delete(&boundlist, index);
    set_widget_value(args[0], "");
  }
  else {
    fprintf(stderr, "mysql: widget %s is not bound\n", args[0]);
  }

  return NULL;
}

char *msp_getbindingenabled(const char *sender) {
  if(bindingenabled)
    return strdup("true");
  else
    return strdup("false");
  return NULL;
}

int msp_setbindingenabled(const char *sender, const char *value) {
  int lastenabled;
  lastenabled = bindingenabled;
  bindingenabled = strtobool(value);
  if(lastenabled == 0 && bindingenabled == 1 && res_set != NULL && boundlist.count > 0) {
    /* Binding has just been re-enabled and there's a result set, so update the widgets */
    update_bound_widgets();
  }
  return 0;
}

void on_scroll(void) {
  /* Check if we need to update bound widgets */
  if(bindingenabled == 1 && boundlist.count > 0)
    update_bound_widgets();

  fireEvent("mysql", "OnScroll");
}

void update_bound_widgets(void) {
  /* Iterate through bound widgets and update each one */
  struct llistitem *ptr;
  int numfields;
  int i;

  numfields = mysql_num_fields(res_set);

  ptr = boundlist.first;
  while(ptr) {
    for(i=0; i < numfields; i++) {
      if(strcasecmp(ptr->value, fields[i].name) == 0) {
        set_widget_value(ptr->name, currentrow[i]);
        break;
      }
    }
    ptr = ptr->next;
  }
}

void set_widget_value(const char *name, const char *value) {
  const char *type;
  type = typeOfObject(name);
  if(type) {
    if(strcmp(type, "textfield") == 0)
      setPropertyAsString(name, "text", value);
    else if(strcmp(type, "label") == 0)
      setPropertyAsString(name, "caption", value);
    else if(strcmp(type, "picture") == 0)
      setPropertyAsString(name, "image", value);
    else if(strcmp(type, "togglebutton") == 0)
      setPropertyAsString(name, "checked", value);
    else
      fprintf(stderr, "mysql: widget %s cannot be data-bound - type %s not supported", name, type);
  }
}


/* Add an item to a linked list */
void llist_add(struct llist *list, const char *name, const char *value) {
  if(list->first == NULL) {
    list->first = malloc(sizeof(struct llistitem));
    list->last = list->first;
  }
  else {
    list->last->next = malloc(sizeof(struct llistitem));
    list->last = list->last->next;
  }
  list->last->name = strdup(name);
  list->last->value = strdup(value);
  list->last->next = NULL;
  list->count++;
}

/* Get an item from a linked list at a specified index */
const char *llist_get(struct llist *list, int index) {
  int i;
  struct llistitem *ptr;

  ptr = list->first;
  for(i=0; ptr!=NULL; i++) {
    if(i==index)
      return ptr->value;
    ptr = ptr->next;
  }

  return NULL;
}

/* Delete an item from a linked list at a specified index */
void llist_delete(struct llist *list, int index) {
  int i;
  struct llistitem *ptr;
  struct llistitem *lastptr;

  ptr = list->first;
  lastptr = NULL;
  for(i=0; ptr!=NULL; i++) {
    if(i==index) {
      if(ptr == list->first)
        list->first = ptr->next;
      if(ptr == list->last)
        list->last = lastptr;
      if(lastptr)
        lastptr->next = ptr->next;
      free(ptr->name);
      free(ptr->value);
      free(ptr);
      list->count--;
      break;
    }
    lastptr = ptr;
    ptr = ptr->next;
  }
}

/* Clear out entire linked list contents */
void llist_clear(struct llist *list) {
  struct llistitem *ptr;
  struct llistitem *freeptr;

  ptr = list->first;
  while(ptr) {
    freeptr = ptr;
    ptr = ptr->next;
    free(freeptr->name);
    free(freeptr->value);
    free(freeptr);
  }
  list->count = 0;
  list->first = NULL;
  list->last = NULL;
}

/* Return the index of an item in a linked list */
int llist_find(struct llist *list, const char *name) {
  int i;
  struct llistitem *ptr;

  ptr = list->first;
  for(i=0; ptr!=NULL; i++) {
    if(strcasecmp(name, ptr->name) == 0)
      return i;
    ptr = ptr->next;
  }

  return -1;
}

/* Convert an integer to a string */
char *inttostr(int n) {
  char *buf;
  buf = malloc(16);
  snprintf(buf, 16, "%d", n);
  return buf;
}

/* Convert a string value into a boolean */
int strtobool(const char *str) {
  if(strcasecmp(str, "true") == 0 || strcasecmp(str, "on") == 0 || strcasecmp(str, "yes") == 0 || atoi(str) != 0)
    return 1;
  else
    return 0;
}
