/***************************************************************************
                  playlist.c - eboxy playlist builder plugin
                             -------------------
    begin                : Sun Oct 13 2002
    copyright            : (C) 2002 by Paul Eggleton
    email                : bluelightning@bluelightning.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>

#include "eboxyplugin.h"
#include "pluginconstants.h"

/* Types */
struct llist {
  char *value;
  struct llist *next;
};

/* Globals */
struct llist *playllist;
struct llist *playllistlast;
int listcount;
char *listboxname;

/* Function prototypes */
char *plb_setup(const char *sender, int numargs, const char *args[]);
char *plb_add(const char *sender, int numargs, const char *args[]);
char *plb_remove(const char *sender, int numargs, const char *args[]);
char *plb_clear(const char *sender, int numargs, const char *args[]);
char *plb_printlist(const char *sender, int numargs, const char *args[]);
char *plb_savelist(const char *sender, int numargs, const char *args[]);
char *plb_getselectedfile(const char *sender);
char *plb_getcount(const char *sender);
int plb_listchoose(const char *sender);

const char *getfilename(const char *path);
const char *getlistitem(int index);
char *inttostr(int n);

/* Implementation */

int ebplugin_init(void) {
  int rc = 0;

  /* Set version info for the plugin */
  setPluginInfo("Playlist Builder", "0.1");

  /* Register an object to be used in scripts */
  rc = registerObject("playlist");
  if(rc)
    return rc;
  rc = registerMethodDL("playlist", "setup", 1, "plb_setup");
  if(rc)
    return rc;
  rc = registerMethodDL("playlist", "add", 1, "plb_add");
  if(rc)
    return rc;
  rc = registerMethodDL("playlist", "remove", 1, "plb_remove");
  if(rc)
    return rc;
  rc = registerMethodDL("playlist", "clear", 0, "plb_clear");
  if(rc)
    return rc;
  rc = registerMethodDL("playlist", "printlist", 0, "plb_printlist");
  if(rc)
    return rc;
  rc = registerMethodDL("playlist", "savelist", 1, "plb_savelist");
  if(rc)
    return rc;
  rc = registerPropertyDL("playlist", "selectedfile", "plb_getselectedfile", NULL);
  if(rc)
    return rc;
  rc = registerPropertyDL("playlist", "count", "plb_getcount", NULL);
  if(rc)
    return rc;

  listboxname = NULL;
  playllist = NULL;
  playllistlast = NULL;
  listcount = 0;

  /* Success */
  return 0;
}

int ebplugin_message(int msgcode, void *msgdata) {
  if(msgcode == PLMSG_BEFOREPAGECHANGE) {
    /* Page is changing */
    if(listboxname) {
      free(listboxname);
      listboxname = NULL;
    }
  }
  return 0;     /* return 0 here for future compatibility */
}

void ebplugin_deinit(void) {
  plb_clear("", 0, NULL);
  if(listboxname)
    free(listboxname);
}

char *plb_setup(const char *sender, int numargs, const char *args[]) {
  listboxname = strdup(args[0]);
  registerEventHandlerDL(listboxname, "OnChoose", "plb_listchoose");

  plb_clear("", 0, NULL);

  return NULL;
}

char *plb_add(const char *sender, int numargs, const char *args[]) {
  const char *margs[1];

  if(playllist == NULL) {
    playllist = malloc(sizeof(struct llist));
    playllistlast = playllist;
  }
  else {
    playllistlast->next = malloc(sizeof(struct llist));
    playllistlast = playllistlast->next;
  }
  playllistlast->value = strdup(args[0]);
  playllistlast->next = NULL;
  listcount++;
  if(listboxname) {
    margs[0] = getfilename(args[0]);
    callMethodNoReturn(listboxname, "additem", 1, margs);
  }

  return NULL;
}

char *plb_remove(const char *sender, int numargs, const char *args[]) {
  struct llist *ptr;
  struct llist *lastptr;
  int i;
  int index;
  char *res;

  index = atoi(args[0]);

  ptr = playllist;
  lastptr = NULL;
  i=0;
  while(ptr) {
    if(i==index) {
      // Remove this item
      if(ptr == playllist)
        playllist = ptr->next;
      if(ptr == playllistlast)
        playllistlast = lastptr;
      if(lastptr)
        lastptr->next = ptr->next;
      free(ptr->value);
      free(ptr);
      listcount--;
      if(listboxname) {
        // Remove item from listbox
        callMethodNoReturn(listboxname, "removeitem", 1, args);
        break;
      }
    }
    i++;
    lastptr = ptr;
    ptr = ptr->next;
  }

  return NULL;
}

char *plb_clear(const char *sender, int numargs, const char *args[]) {
  struct llist *ptr;
  struct llist *freeptr;
  char *res;

  ptr = playllist;
  while(ptr) {
    freeptr = ptr;
    ptr = ptr->next;
    free(freeptr->value);
    free(freeptr);
  }
  listcount = 0;
  playllist = NULL;
  playllistlast = NULL;

  if(listboxname) {
    /* Clear listbox */
    callMethodNoReturn(listboxname, "clear", 0, NULL);
  }

  return NULL;
}

char *plb_printlist(const char *sender, int numargs, const char *args[]) {
  struct llist *ptr;

  ptr = playllist;
  while(ptr) {
    printf("%s\n", ptr->value);
    ptr = ptr->next;
  }

  return NULL;
}

char *plb_savelist(const char *sender, int numargs, const char *args[]) {
  struct llist *ptr;
  FILE *fp;

  fp = fopen(args[0], "w");
  if(fp) {
    ptr = playllist;
    while(ptr) {
      fprintf(fp, "%s\n", ptr->value);
      ptr = ptr->next;
    }
    fclose(fp);
  }
  else
    fprintf(stderr, "Failed to open %s for writing\n", args[0]);

  return NULL;
}

char *plb_getselectedfile(const char *sender) {
  const char *selectedidx;
  int index;

  if(listboxname) {
    selectedidx = getPropertyAsString(listboxname, "selectedindex");
    index = atoi(selectedidx);
    if(index > -1 && index < listcount)
      return strdup(getlistitem(index));
  }
  
  return NULL;
}

char *plb_getcount(const char *sender) {
  return inttostr(listcount);
}

/* Called whenever the user chooses an item from the list */
int plb_listchoose(const char *sender) {
  /* A file was chosen */
  fireEvent("playlist", "OnChooseFile");

  return 0;
}

const char *getfilename(const char *path) {
  int i;

  for(i=strlen(path)-1; i>=0; i--) {
    if(path[i]=='/') {
      return path+i+1;
    }
  }
  return path;
}

const char *getlistitem(int index) {
  int i;
  struct llist *ptr;

  ptr = playllist;
  for(i=0; ptr!=NULL; i++) {
    if(i==index)
      return ptr->value;
    ptr = ptr->next;
  }
  return "";
}

/* Convert an integer to a string */
char *inttostr(int n) {
  char *buf;
  buf = (char *)malloc(16);
  snprintf(buf, 16, "%d", n);
  return buf;
}

