/***************************************************************************
          xine.c - eboxy multimedia player plugin utilising xine-lib
                          -------------------
    begin      : Thu Jun 26 2003
    copyright  : Copyright (C) 2003 by Paul Eggleton <bluelightning@bluelightning.org>
                 portions Copyright (C) 2003 Daniel Caujolle-Bert <segfault@club-internet.fr>
                 (initially based on example code from the xine hackers' guide)
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include <unistd.h>
#include <errno.h>
#include <signal.h>
#include <pthread.h>

#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/Xatom.h>
#include <X11/Xutil.h>
#include <X11/extensions/XShm.h>

#include <xine.h>
#include <xine/xineutils.h>

#include "SDL.h"
#include "SDL_syswm.h"


#include "eboxyplugin.h"
#include "pluginconstants.h"


/* Globals */
static xine_t              *xine;
static xine_stream_t       *stream;
static xine_video_port_t   *vo_port;
static xine_audio_port_t   *ao_port;
static xine_event_queue_t  *event_queue;
static x11_visual_t        vis;

static pthread_t           x11_event_thread;
pthread_mutex_t            deinit_mutex;
pthread_cond_t             deinit_cond;

static Display             *display;
static int                  screen;
static Window               window[2];
static int                  completion_event;
static int                  xpos, ypos, width, height, origwidth, origheight, fullscreen;
static double               pixel_aspect;
static SDL_SysWMinfo        info;
static char                *mrl;

static int                  running = 0;


/* Function prototypes */
char *exin_play(const char *sender, int numargs, const char *args[]);
char *exin_stop(const char *sender, int numargs, const char *args[]);
char *exin_pause(const char *sender, int numargs, const char *args[]);
char *exin_togglefullscreen(const char *sender, int numargs, const char *args[]);
int exin_setmrl(const char *sender, const char *value);
char *exin_getmrl(const char *sender);
int exin_setx(const char *sender, const char *value);
char *exin_getx(const char *sender);
int exin_sety(const char *sender, const char *value);
char *exin_gety(const char *sender);
int exin_setwidth(const char *sender, const char *value);
char *exin_getwidth(const char *sender);
int exin_setheight(const char *sender, const char *value);
char *exin_getheight(const char *sender);

int new_thread(void *(*thread_proc)(void *));
void *x11_eventloop_thread_proc(void *arg);

int init_xine(void);
int init_output(void);
void deinit_xine(void);
void deinit_output(void);
void signal_handler(int sig);
void update_output_pos(void);
void update_output_size(void);
void send_mouse_event(int x, int y, int button);

void eXHideBorder(Display *disp, Window *wnd);
char *inttostr(int n);

#ifndef XShmGetEventBase
extern int XShmGetEventBase(Display *);
#endif

#define MWM_HINTS_DECORATIONS   (1L << 1)
#define PROP_MWM_HINTS_ELEMENTS 5
typedef struct {
  uint32_t  flags;
  uint32_t  functions;
  uint32_t  decorations;
  int32_t   input_mode;
  uint32_t  status;
} MWMHints;


/* Implementation */

int ebplugin_init(void) {
  int rc = 0;

  /* Set version info for the plugin */
  setPluginInfo("eboxy xine plugin", "0.1");

  /* Register an object to be used in scripts */
  rc = registerObject("xine");
  if(rc)
    return rc;
  rc = registerMethodDL("xine", "play", 0, "exin_play");
  if(rc)
    return rc;
  rc = registerMethodDL("xine", "stop", 0, "exin_stop");
  if(rc)
    return rc;
  rc = registerMethodDL("xine", "pause", 0, "exin_pause");
  if(rc)
    return rc;
  rc = registerMethodDL("xine", "togglefullscreen", 0, "exin_togglefullscreen");
  if(rc)
    return rc;
  rc = registerPropertyDL("xine", "x", "exin_getx", "exin_setx");
  if(rc)
    return rc;
  rc = registerPropertyDL("xine", "y", "exin_gety", "exin_sety");
  if(rc)
    return rc;
  rc = registerPropertyDL("xine", "width", "exin_getwidth", "exin_setwidth");
  if(rc)
    return rc;
  rc = registerPropertyDL("xine", "height", "exin_getheight", "exin_setheight");
  if(rc)
    return rc;
  rc = registerPropertyDL("xine", "mrl", "exin_getmrl", "exin_setmrl");
  if(rc)
    return rc;

  /* Initialise sync objects for deinitialisation */
  pthread_mutex_init(&deinit_mutex, NULL);
  pthread_cond_init(&deinit_cond, NULL);

  /* Initialise X11 output */
  if(init_output())
    return -1;
  /* Initialise xine library */
  if(init_xine()) {
    deinit_output();
    return -1;
  }
  /* Create a thread to handle events in the X11 window */
  if(new_thread(x11_eventloop_thread_proc)) {
    deinit_xine();
    deinit_output();
    return -1;
  }

  /* Success */
  return 0;
}

int ebplugin_message(int msgcode, void *msgdata) {
  return 0;     /* return 0 here for future compatibility */
}

void ebplugin_deinit(void) {
  void *threadret;

  running = 0;

  if(mrl)
    free(mrl);
  deinit_xine();
  deinit_output();
  /* Wait for thread */
  pthread_join(x11_event_thread, &threadret);
}

char *exin_play(const char *sender, int numargs, const char *args[]) {
  /* FIXME: play after pause should not restart media */
  if(mrl != NULL && strlen(mrl) > 0) {
    if((!xine_open(stream, mrl)) || (!xine_play(stream, 0, 0))) {
      printf("Unable to open mrl '%s'\n", mrl);
      return NULL;
    }
  }

  return NULL;
}

char *exin_stop(const char *sender, int numargs, const char *args[]) {
  if(xine_get_status(stream) == XINE_STATUS_PLAY) {
    xine_stop(stream);
  }
  return NULL;
}

char *exin_pause(const char *sender, int numargs, const char *args[]) {
  if(xine_get_status(stream) == XINE_STATUS_PLAY) {
    if(xine_get_param(stream, XINE_PARAM_SPEED) == XINE_SPEED_PAUSE)
      xine_set_param(stream, XINE_PARAM_SPEED, XINE_SPEED_NORMAL);
    else
      xine_set_param(stream, XINE_PARAM_SPEED, XINE_SPEED_PAUSE);
  }
  return NULL;
}

char *exin_togglefullscreen(const char *sender, int numargs, const char *args[]) {
  Window    tmp_win;

  if(!fullscreen) {
    origwidth = width;
    origheight = height;
  }

  XLockDisplay(display);
  XUnmapWindow(display, window[fullscreen]);
  fullscreen = !fullscreen;

  if(!fullscreen) {
    width = origwidth;
    height = origheight;
  }

  XMapRaised(display, window[fullscreen]);
  XSync(display, False);
  XTranslateCoordinates(display, window[fullscreen],
                        DefaultRootWindow(display),
                        0, 0, &xpos, &ypos, &tmp_win);
  XUnlockDisplay(display);

  xine_gui_send_vo_data(stream, XINE_GUI_SEND_DRAWABLE_CHANGED,
                        (void*) window[fullscreen]);

  return NULL;
}

char *exin_getmrl(const char *sender) {
  if(mrl)
    return strdup(mrl);
  else
    return NULL;
}

int exin_setmrl(const char *sender, const char *value) {
  mrl = strdup(value);
  return 0;
}

int exin_setx(const char *sender, const char *value) {
  int intval = atoi(value);
  if(intval > 0)
    xpos = intval;
  update_output_pos();

  return 0;
}

char *exin_getx(const char *sender){
  return inttostr(xpos);
}

int exin_sety(const char *sender, const char *value){
  int intval = atoi(value);
  if(intval > 0)
    ypos = intval;
  update_output_pos();

  return 0;
}

char *exin_gety(const char *sender){
  return inttostr(ypos);
}

int exin_setwidth(const char *sender, const char *value){
  int intval = atoi(value);
  if(intval > 0)
    width = intval;
  update_output_size();

  return 0;
}

char *exin_getwidth(const char *sender){
  return inttostr(width);
}

int exin_setheight(const char *sender, const char *value){
  int intval = atoi(value);
  if(intval > 0)
    height = intval;
  update_output_size();

  return 0;
}

char *exin_getheight(const char *sender){
  return inttostr(height);
}


/* Callback function to tell xine what video output size to use */
static void dest_size_cb(void *data, int video_width, int video_height, double video_pixel_aspect,
                         int *dest_width, int *dest_height, double *dest_pixel_aspect)  {

  if(!running)
    return;

  *dest_width        = width;
  *dest_height       = height;
  *dest_pixel_aspect = pixel_aspect;
}

/* Callback function to tell xine what video output parameters to use */
static void frame_output_cb(void *data, int video_width, int video_height,
                            double video_pixel_aspect, int *dest_x, int *dest_y,
                            int *dest_width, int *dest_height,
                            double *dest_pixel_aspect, int *win_x, int *win_y) {
  if(!running)
    return;

  *dest_x            = 0;
  *dest_y            = 0;
  *win_x             = xpos;
  *win_y             = ypos;
  *dest_width        = width;
  *dest_height       = height;
  *dest_pixel_aspect = pixel_aspect;
}

/* Callback function for xine events */
static void event_listener(void *user_data, const xine_event_t *event) {
  switch(event->type) {
  case XINE_EVENT_UI_PLAYBACK_FINISHED:
    {
      fireEvent("xine", "OnFinish");
      break;
    }

  case XINE_EVENT_PROGRESS:
    {
      xine_progress_data_t *pevent = (xine_progress_data_t *) event->data;

      printf("%s [%d%%]\n", pevent->description, pevent->percent);
    }
    break;

  }
}

int init_output(void) {
  double            res_h, res_v;

  if(!XInitThreads()) {
    printf("XInitThreads() failed\n");
    return 1;
  }

  SDL_VERSION(&info.version);
  if( SDL_GetWMInfo(&info) ) {
    if( info.subsystem != SDL_SYSWM_X11 ) {
      fprintf(stderr, "eboxy is not running under X11 - this plugin requires X\n");
      return -1;
    }
  }
        
  display = XOpenDisplay(NULL);

  screen  = XDefaultScreen(display);
  xpos    = 0;
  ypos    = 0;
  width   = 320;
  height  = 200;
  origwidth = width;
  origheight = height;
  

  XLockDisplay(display);
  fullscreen = 0;
  window[0] = XCreateSimpleWindow(display, /*XDefaultRootWindow(display)*/ info.info.x11.window,
                                  xpos, ypos, width, height, 1, 0, 0);
  window[1] = XCreateSimpleWindow(display, XDefaultRootWindow(display),
                                  0, 0, (DisplayWidth(display, screen)),
                                  (DisplayHeight(display, screen)), 0, 0, 0);

  XSelectInput(display, window[0], (ExposureMask | ButtonPressMask /*| KeyPressMask*/ |
                                    /*ButtonMotionMask |*/ StructureNotifyMask |
                                    PropertyChangeMask | PointerMotionMask));

  XSelectInput(display, window[1], (ExposureMask | ButtonPressMask | KeyPressMask | KeyReleaseMask |
                                    ButtonMotionMask | StructureNotifyMask |
                                    PropertyChangeMask | PointerMotionMask));

  eXHideBorder(display, &window[0]);
  eXHideBorder(display, &window[1]);

  if(XShmQueryExtension(display) == True)
    completion_event = XShmGetEventBase(display) + ShmCompletion;
  else
    completion_event = -1;

  XMapRaised(display, window[fullscreen]);

  res_h = (DisplayWidth(display, screen) * 1000 / DisplayWidthMM(display, screen));
  res_v = (DisplayHeight(display, screen) * 1000 / DisplayHeightMM(display, screen));
  XSync(display, False);
  XUnlockDisplay(display);

  vis.display           = display;
  vis.screen            = screen;
  vis.d                 = window[fullscreen];
  vis.dest_size_cb      = dest_size_cb;
  vis.frame_output_cb   = frame_output_cb;
  vis.user_data         = NULL;
  pixel_aspect          = res_v / res_h;

  if(fabs(pixel_aspect - 1.0) < 0.01)
    pixel_aspect = 1.0;

  return 0;
}

int init_xine(void) {
  char              configfile[2048];
  char             *vo_driver = "auto";
  char             *ao_driver = "auto";

  mrl = NULL;

  xine = xine_new();
  sprintf(configfile, "%s%s", xine_get_homedir(), "/.xine/config2");
  xine_config_load(xine, configfile);
  xine_init(xine);

  vo_port = xine_open_video_driver(xine, vo_driver, XINE_VISUAL_TYPE_X11, (void *) &vis);
  ao_port = xine_open_audio_driver(xine , ao_driver, NULL);

  stream = xine_stream_new(xine, ao_port, vo_port);
  event_queue = xine_event_new_queue(stream);
  xine_event_create_listener_thread(event_queue, event_listener, NULL);

  xine_gui_send_vo_data(stream, XINE_GUI_SEND_DRAWABLE_CHANGED, (void *) window[fullscreen]);
  xine_gui_send_vo_data(stream, XINE_GUI_SEND_VIDEOWIN_VISIBLE, (void *) 1);

  running = 1;

  return 0;
}

void deinit_xine(void) {
  xine_close(stream);
  xine_event_dispose_queue(event_queue);
  xine_dispose(stream);
  xine_close_audio_driver(xine, ao_port);
  xine_close_video_driver(xine, vo_port);
  xine_exit(xine);
}

void deinit_output(void) {
  XLockDisplay(display);
  XUnmapWindow(display,  window[fullscreen]);
  XDestroyWindow(display,  window[0]);
  XDestroyWindow(display,  window[1]);
  XUnlockDisplay(display);

  XCloseDisplay (display); 
}

int new_thread(void *(*thread_proc)(void *)) {
  pthread_attr_t thread_attr;
  int res;

  /* Need to start the thread up in the detached state, or we run out of thread IDs
     fast! */
  res = pthread_attr_init(&thread_attr);
  if(res) {
    fprintf(stderr, "Failed to create thread attribute (%d)", res);
    return -1;
  }
  res = pthread_attr_setdetachstate(&thread_attr, PTHREAD_CREATE_DETACHED);
  if(res) {
    fprintf(stderr, "Failed to set thread detached state (%d)", res);
    return -1;
  }
  /* Now create the thread */
  res = pthread_create(&x11_event_thread, &thread_attr, thread_proc, (void *) 0);
  if(res) {
    fprintf(stderr, "Failed to create event thread (%d)", res);
    return -1;
  }

  /* Done with the attribute, release it */
  pthread_attr_destroy(&thread_attr);
  return 0;
}

void *x11_eventloop_thread_proc(void *arg) {
  (void) signal(SIGPIPE, signal_handler);

  while(running) {
    XEvent   xevent;

    XNextEvent(display, &xevent);

    if(running == 0)
      break; /* probably deinitialisation time */
    
    switch(xevent.type) {

    case KeyRelease:
    case KeyPress:
      {
        XKeyEvent  kevent;
        kevent = xevent.xkey;

        /* Pass key events back to eboxy when in full-screen mode (yes this is a hack...) */
        info.info.x11.lock_func();
        
        XSendEvent(info.info.x11.display, info.info.x11.window, True, KeyPressMask, (XEvent *) &kevent);
        info.info.x11.unlock_func();
        break;
      }
      break;

    case MotionNotify:
      {
        XMotionEvent  mevent;
        mevent = xevent.xmotion;

        send_mouse_event(mevent.x, mevent.y, 0);
      }
      break;

    case ButtonPress:
      {
        XButtonEvent  bevent;
        int button;
        bevent = xevent.xbutton;

        if(bevent.button == Button1)
          button = 1;
        else if(bevent.button == Button2)
          button = 2;
        else if(bevent.button == Button3)
          button = 3;
        else
          break;
          
        send_mouse_event(0, 0, button);
      }
      break;
      
    case Expose:
      if(xevent.xexpose.count != 0)
        break;
      xine_gui_send_vo_data(stream, XINE_GUI_SEND_EXPOSE_EVENT, &xevent);
      break;

    /* This happens when the output window is resized/moved */
    case ConfigureNotify:
      {
        XConfigureEvent *cev = (XConfigureEvent *) &xevent;
        Window           tmp_win;

        if(cev->window == window[fullscreen]) {
          width  = cev->width;
          height = cev->height;

          if((cev->x == 0) && (cev->y == 0)) {
            XLockDisplay(display);
            XTranslateCoordinates(display, cev->window,
                                  DefaultRootWindow(cev->display),
                                  0, 0, &xpos, &ypos, &tmp_win);
            XUnlockDisplay(display);
          }
          else {
            xpos = cev->x;
            ypos = cev->y;
          }
        }
      }
      break;

    }

    if(xevent.type == completion_event)
      xine_gui_send_vo_data(stream, XINE_GUI_SEND_COMPLETION_EVENT, &xevent);
  }

/* FIXME  requestUnload(); */
  pthread_exit((void *) 0);
}

void signal_handler(int sig) {
  switch(sig) {
    case SIGPIPE: /* This doesn't seem to work */
      fprintf(stderr, "WARNING! Video window closed\n");
      running = 0;
      (void) signal(sig, SIG_IGN);
      break;
    default:
      (void) signal(sig, SIG_DFL);
  }
}

void update_output_pos(void) {
  XLockDisplay(display);
  XMoveWindow(display, window[0], xpos, ypos);
  XSync(display, False);
  XUnlockDisplay(display);
}

void update_output_size(void) {
  XLockDisplay(display);
  XResizeWindow(display, window[0], width, height);
  XSync(display, False);
  XUnlockDisplay(display);
}

void send_mouse_event(int x, int y, int button) {
  xine_event_t xmevent;
  xine_input_data_t xmdata;
  static int tr_x=0, tr_y=0;
  x11_rectangle_t rect;

  rect.x = x;
  rect.y = y;
  rect.w = rect.h = 0;
  if(xine_gui_send_vo_data(stream, XINE_GUI_SEND_TRANSLATE_GUI_TO_VIDEO, (void *)&rect) != -1) {
    if(button == 0) {
      xmevent.type = XINE_EVENT_INPUT_MOUSE_MOVE;
      xmdata.x = rect.x;
      xmdata.y = rect.y;
      tr_x = rect.x;
      tr_y = rect.y;
    }
    else {
      xmevent.type = XINE_EVENT_INPUT_MOUSE_BUTTON;
      xmdata.button = button;
      xmdata.x = tr_x;
      xmdata.y = tr_y;
    }

    xmevent.stream = stream;
    xmevent.data = &xmdata;
    xmevent.data_length = sizeof(xmdata);

    xine_event_send(stream, &xmevent);
  }
}


void eXHideBorder(Display *disp, Window *wnd) {
  Atom              XA_NO_BORDER;
  MWMHints          mwmhints;

  XA_NO_BORDER         = XInternAtom(display, "_MOTIF_WM_HINTS", False);
  mwmhints.flags       = MWM_HINTS_DECORATIONS;
  mwmhints.decorations = 0;
  XChangeProperty(disp, *wnd,
                  XA_NO_BORDER, XA_NO_BORDER, 32, PropModeReplace, (unsigned char *) &mwmhints,
                  PROP_MWM_HINTS_ELEMENTS);
}

/* Convert an integer to a string */
char *inttostr(int n) {
  char *buf;
  buf = malloc(16);
  snprintf(buf, 16, "%d", n);
  return buf;
}

/* Convert a string value into a boolean */
int strtobool(const char *str) {
  if(strcasecmp(str, "true") == 0 || strcasecmp(str, "on") == 0 || strcasecmp(str, "yes") == 0 || atoi(str) != 0)
    return 1;
  else
    return 0;
}

/* Convert boolean into a string value */
const char *booltostr(int value) {
  if(value)
    return "true";
  else
    return "false";
}
